#!/usr/bin/env python3
import argparse
import logging
import os

from pogoraidbot import PoGORaidBot
from .log import logger_setup

if __name__ == "__main__":
    # Gets inline arguments
    parser = argparse.ArgumentParser(prog="pogoraidbot")

    parser.add_argument("-t", "--token", dest="token", help="telegram bot token")
    parser.add_argument("-r", "--redis", dest="redis", help="redis url in \"redis://{host}[:port]/{db}\" format")
    parser.add_argument("-a", "--superadmin", dest="superadmin", help="superadmin's id")
    parser.add_argument("-b", "--bosses-file", dest="bosses_file",
                        help="JSON or CSV file contains possible pokémons in the raids. It can be also provided over http(s)")
    parser.add_argument("-o", "--bosses-expiration", dest="bosses_expiration",
                        help="Validity of the bosses list in hours")
    parser.add_argument("-g", "--gyms-file", dest="gyms_file",
                        help="JSON file contains gyms and their coordinates. It can be also provided over http(s)")
    parser.add_argument("-y", "--gyms-expiration", dest="gyms_expiration",
                        help="Validity of the gyms list in hours")
    parser.add_argument("-e", "--env", dest="env", action="store_true",
                        help="Use environment variables for the configuration")
    parser.add_argument("-d", "--debug-folder", dest="debug_folder", help="debug folder")
    parser.add_argument("-v", dest="log_level", action="count",
                        help="number of -v specifics level of verbosity")
    parser.add_argument("--info", dest="log_level", action="store_const", const=2, help="equal to -vv")
    parser.add_argument("--debug", dest="log_level", action="store_const", const=3, help="equal to -vvv")

    # Parses args
    args = vars(parser.parse_args())

    # Removes None elements
    args = {k: args[k] for k in args if args[k] is not None}

    if args["env"]:
        env = {
            "token": os.getenv("PGRB_BOT_TOKEN"),
            "redis": os.getenv("PGRB_BOT_REDIS"),
            "superadmin": os.getenv("PGRB_BOT_SUPERADMIN"),
            "gyms_file": os.getenv("PGRB_BOT_GYMS_FILE"),
            "gyms_expiration": os.getenv("PGRB_BOT_GYMS_EXPIRATION"),
            "bosses_file": os.getenv("PGRB_BOT_BOSSES_FILE"),
            "bosses_expiration": os.getenv("PGRB_BOT_BOSSES_EXPIRATION"),
            "log_level": os.getenv("PGRB_BOT_LOG_LEVEL")
        }

        if os.getenv("PGRB_BOT_DEBUG_PATH") is not None:
            env["debug_folder"] = "/srv"

        # Removes None elements
        env = {k: env[k] for k in env if env[k] is not None}

        args = {**env, **args}

    del args["env"]

    # Parses the verbosity level
    try:
        logger_setup({
                         0: logging.ERROR,
                         1: logging.WARNING,
                         2: logging.INFO,
                         3: logging.DEBUG,
                         "ERROR": logging.ERROR,
                         "WARNING": logging.WARNING,
                         "INFO": logging.INFO,
                         "DEBUG": logging.DEBUG
                     }[args["log_level"]])

    except KeyError:
        logger_setup()

    if "log_level" in args:
        del args["log_level"]

    # Creates the bot
    bot = PoGORaidBot(**args)

    # Manages updates
    bot.listen()



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Union, List

from .data import DataList, Data
from .exceptions import InvalidJSON

gyms: Union[List[Gym], None] = None


@dataclass
class Gym(Data):
    latitude: float = None
    longitude: float = None

    @property
    def map(self):
        if self.latitude is None or self.longitude is None:
            return None

        return "https://maps.google.com/maps?q={latitude},{longitude}&ll={latitude},{longitude}&z=17".format(
            **self.__dict__)


class GymsList(DataList):
    def _load_json(self, raw: str) -> None:
        try:
            data = json.loads(raw)
        except ValueError:
            raise InvalidJSON

        # Try to find a list of gyms in first level of the file
        gyms_raw = None
        for _, l in data.items():
            if isinstance(l, list):
                if isinstance(l[0], dict):
                    if "name" in l[0] and "latitude" in l[0] and "longitude" in l[0]:
                        gyms_raw = l
                        break

        # Check if a list is found
        if gyms_raw is None:
            raise InvalidJSON

        self.clear()

        # Add each gyms to the list
        for g in gyms_raw:
            self.append(Gym(g["name"], g["latitude"], g["longitude"]))



import functools
from typing import Callable


class CachedMethod:
    def __init__(self, func: Callable):
        self.func = func

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)

    def __call__(self, *args, **kwargs):
        inst = args[0]
        func_name = self.func.__name__

        try:
            inst.__cache__
        except AttributeError:
            inst.__cache__ = {}

        try:
            return inst.__cache__[func_name]
        except KeyError:
            pass

        inst.__cache__[func_name] = self.func(*args, **kwargs)

        return inst.__cache__[func_name]



from __future__ import annotations

import datetime
import random
import string
from dataclasses import dataclass, field
from enum import Enum
from functools import reduce
from typing import Dict

from telegram import User
from telegram.utils import helpers

from ..data import Boss, Gym


@dataclass
class Participant:
    class Type(Enum):
        NORMAL = 1
        REMOTE = 2
        REMOTE_INVITE = 3
        FLYER = 4

    id: int
    name: str
    type: Type = Type.NORMAL
    number: int = 1

    @property
    def is_remote(self) -> bool:
        return self.type == Participant.Type.REMOTE

    @property
    def is_remote_invite(self) -> bool:
        return self.type == Participant.Type.REMOTE_INVITE

    @property
    def is_flyer(self) -> bool:
        return self.type == Participant.Type.FLYER


@dataclass
class Raid:
    code: str = field(
        default_factory=lambda: "".join(random.choice(string.ascii_letters + string.digits) for i in range(8)))
    gym: Gym = None
    is_ex: bool = False
    level: int = None
    is_hatched: bool = False
    end: datetime.time = None
    hatching: datetime.time = None
    hangout: datetime.time = None
    boss: Boss = None
    is_aprx_time: bool = False
    participants: Dict[int, Participant] = field(default_factory=lambda: {})

    def add_participant(self, user: User) -> None:
        if user.id in self.participants:
            self.participants[user.id].name = user.full_name
            self.participants[user.id].number += 1
        else:
            self.participants[user.id] = Participant(user.id, user.full_name)

    def remove_participant(self, user: User) -> bool:
        if user.id in self.participants:
            if self.participants[user.id].number > 1:
                self.participants[user.id].name = user.full_name
                self.participants[user.id].number -= 1
            else:
                del self.participants[user.id]
            return True
        return False

    def toggle_remote(self, user: User) -> bool:
        if user.id in self.participants:
            self.participants[user.id].name = user.full_name
            self.participants[user.id].type = Participant.Type.REMOTE if self.participants[
                                                                             user.id].type != Participant.Type.REMOTE else Participant.Type.NORMAL
            return True
        return False

    def toggle_remote_invite(self, user: User) -> bool:
        if user.id in self.participants:
            self.participants[user.id].name = user.full_name
            self.participants[user.id].type = Participant.Type.REMOTE_INVITE if self.participants[
                                                                             user.id].type != Participant.Type.REMOTE_INVITE else Participant.Type.NORMAL
            return True
        return False

    def toggle_flyer(self, user: User) -> bool:
        if user.id in self.participants:
            self.participants[user.id].name = user.full_name
            self.participants[user.id].type = Participant.Type.FLYER if self.participants[
                                                                            user.id].type != Participant.Type.FLYER else Participant.Type.NORMAL
            return True
        return False

    @property
    def participants_count(self) -> int:
        return reduce((lambda x, y: x + y), [p.number for _, p in self.participants.items()], 0)

    @property
    def effective_level(self) -> int:
        if self.boss is not None:
            if self.boss.level is not None:
                return self.boss.level
        return self.level

    def to_msg(self) -> str:
        msg = ""

        if self.is_ex:
            msg += "*EX*"

        if self.gym.latitude is not None and self.gym.longitude is not None:
            msg += "[{}]({})".format(
                helpers.escape_markdown(self.gym.name, version=2),
                helpers.escape_markdown(self.gym.map, version=2))
        else:
            msg += "{}".format(helpers.escape_markdown(self.gym.name, version=2))

        msg += "\n\n"

        if self.boss is not None:
            if self.boss.is_there_shiny:
                msg += "\U00002728*{}*\U00002728".format(self.boss.name)
            else:
                msg += "{}".format(self.boss.name)
            msg += "\n"

        if self.effective_level is not None:
            msg += "\U00002B50" * self.effective_level
            msg += "\n"

        msg += "\n"

        if self.hatching is not None:
            msg += "`Hatching:   {}{}`\n".format("~" if self.is_aprx_time else " ", self.hatching.strftime('%H:%M'))

        if self.end is not None:
            msg += "`Ends:       {}{}`\n".format("~" if self.is_aprx_time else " ", self.end.strftime('%H:%M'))

        if self.hangout is not None:
            msg += "`Hangout:     {}`\n".format(self.hangout.strftime('%H:%M'))

        if self.participants_count > 0:
            msg += "`─────────────────────`\n"

            for p_id, p in self.participants.items():
                msg += helpers.mention_markdown(p_id, p.name, version=2)

                if p.is_remote:
                    msg += "\U0001F3E1"
                if p.is_remote_invite:
                    msg += "\U0001F48C"
                if p.is_flyer:
                    msg += "\U00002708"

                if p.number > 1:
                    msg += " \\+{}".format(p.number - 1)

                msg += "\n"

            msg += "`─────────────────────`\n"
            msg += "*{}* participants\n".format(self.participants_count)

        msg += "`[{}]`".format(self.code)

        return msg



from __future__ import annotations

import datetime
import functools
import logging
import os
import pickle
import re
import sys
import traceback
from typing import Callable

import cv2
from apscheduler.schedulers.background import BackgroundScheduler
from redis import StrictRedis, exceptions
from telegram import InlineKeyboardButton, InlineKeyboardMarkup, ParseMode, Update, Bot, Message, error
from telegram.ext import Updater, MessageHandler, CallbackQueryHandler, CommandHandler, CallbackContext
from telegram.ext.filters import Filters

from .. import redis_keys
from ..data import bosses, gyms
from ..raid import Raid
from ..screenshot import ScreenshotRaid

_LOGGER = logging.getLogger(__package__)


class PoGORaidBot:
    class Decorator:
        class ChatMustBeEnabled:
            def __init__(self, func: Callable[[PoGORaidBot, Update, CallbackContext], bool]):
                self.func = func

            def __get__(self, obj, objtype):
                """Support instance methods."""
                return functools.partial(self.__call__, obj)

            def __call__(self, inst: PoGORaidBot, update: Update, context: CallbackContext) -> bool:
                try:
                    chat_id = update.message.chat_id
                except AttributeError:
                    chat_id = update.callback_query.message.chat_id

                # Check if this chat is enabled
                if not inst._redis.sismember(redis_keys.ENABLEDCHAT, chat_id):
                    _LOGGER.info("Chat {} is not enabled".format(chat_id))

                    return False

                return self.func(inst, update, context)

        class UserMustBeAdmin:
            def __init__(self, func: Callable[[PoGORaidBot, Update, CallbackContext], bool]):
                self.func = func

            def __get__(self, obj, objtype):
                """Support instance methods."""
                return functools.partial(self.__call__, obj)

            def __call__(self, inst: PoGORaidBot, update: Update, context: CallbackContext) -> bool:
                is_admin = False
                # If the chat is private doesn't check administrators
                if update.message.chat.type == update.message.chat.PRIVATE:
                    is_admin = True
                else:
                    # Get the list of administrators of a chat
                    for a in context.bot.get_chat_administrators(update.message.chat.id):
                        # Check if the current admin in the user
                        if a.user.id == update.message.from_user.id:
                            is_admin = True
                            break

                # Check if the sender is an admin
                if not is_admin:
                    _LOGGER.info("User {} is not admin".format(update.message.from_user.id))

                    return False

                return self.func(inst, update, context)

        class UserMustBeBotAdmin:
            def __init__(self, func: Callable[[PoGORaidBot, Update, CallbackContext], bool]):
                self.func = func

            def __get__(self, obj, objtype):
                """Support instance methods."""
                return functools.partial(self.__call__, obj)

            def __call__(self, inst: PoGORaidBot, update: Update, context: CallbackContext) -> bool:
                # Check if the user is a bot admin
                if not inst._redis.sismember(redis_keys.ADMIN, update.message.from_user.id):
                    _LOGGER.warning("User {} is not a bot admin".format(update.message.from_user.id))

                    return False

                return self.func(inst, update, context)

    def __init__(self,
                 token: str,
                 redis: str = "redis://127.0.0.1:6379/0",
                 superadmin: int = None,
                 bosses_file: str = None,
                 bosses_expiration: int = 12,
                 gyms_file: str = None,
                 gyms_expiration: int = 12,
                 debug_folder: str = None
                 ):
        # Init and test redis connection
        self._redis = StrictRedis.from_url(url=redis, charset="utf-8", decode_responses=False)

        _LOGGER.info("Try to connect to Redis...")
        try:
            self._redis.ping()
        except exceptions.ConnectionError:
            _LOGGER.critical("Unable to connect to Redis")
            sys.exit()
        _LOGGER.info("Successfully connected to Redis")

        # Save superadmin
        self._superadmin = int(superadmin) if superadmin is not None else None
        # Add superadmin to the admins db
        if self._superadmin is not None:
            self._redis.set(redis_keys.SUPERADMIN, self._superadmin)
            self._redis.sadd(redis_keys.ADMIN, self._superadmin)

        # Save debug folder
        self._debug_folder = debug_folder
        if self._debug_folder is not None:
            self._debug_folder = os.path.abspath(debug_folder)
            ScreenshotRaid.debug = True
            _LOGGER.info("\"{}\" was set as debug folder".format(self._debug_folder))

        # Init the bot
        self._bot = Bot(token)

        # Init updater
        self._updater = Updater(bot=self._bot, use_context=True)

        # Get the id of the bot
        self._id = self._bot.get_me().id

        # Set the handler functions
        # Set the handler for screens
        self._updater.dispatcher.add_handler(MessageHandler(Filters.photo, self._handler_screenshot))
        # Set the handler to set the hangout
        self._updater.dispatcher.add_handler(MessageHandler(
            Filters.reply & Filters.regex(r"^\s*[0-2]?[0-9][:.,][0-5]?[0-9]\s*$"), self._handler_set_hangout))
        # Set the handler for the buttons
        self._updater.dispatcher.add_handler(CallbackQueryHandler(self._handler_buttons))
        # Set the handler for the pinned message notify
        self._updater.dispatcher.add_handler(MessageHandler(Filters.status_update.pinned_message,
                                                            self._handler_event_pinned))
        # Set the handler to set the boss
        self._updater.dispatcher.add_handler(MessageHandler(
            Filters.reply & Filters.regex(r"^\s*[a-zA-Z]+\s*$"), self._handler_set_boss))

        # Set the handler for scan command
        self._updater.dispatcher.add_handler(CommandHandler("scan", self._handler_command_scan))
        # Set the handler for enablechat command
        self._updater.dispatcher.add_handler(CommandHandler("enablechat", self._handler_command_enablechat))
        # Set the handler for disablechat command
        self._updater.dispatcher.add_handler(CommandHandler("disablechat", self._handler_command_disablechat))
        # Set the handler for enablescan command
        self._updater.dispatcher.add_handler(CommandHandler("enablescan", self._handler_command_enablescan))
        # Set the handler for disablescan command
        self._updater.dispatcher.add_handler(CommandHandler("disablescan", self._handler_command_disablescan))
        # Set the handler for addadmin command
        self._updater.dispatcher.add_handler(CommandHandler("addadmin", self._handler_command_addadmin, Filters.reply))
        # Set the handler for removeadmin command
        self._updater.dispatcher.add_handler(CommandHandler("removeadmin", self._handler_command_removeadmin,
                                                            Filters.reply))

        # Set the handler for the errors
        self._updater.dispatcher.add_error_handler(self._handler_error)

        # Creates background scheduler for update the db
        self._scheduler = BackgroundScheduler(daemon=True)

        # Creates job to update bosses list
        if bosses_file is not None:
            bosses.load_from(bosses_file)
            self._scheduler.add_job(lambda: bosses.load_from(bosses_file), 'interval', hours=int(bosses_expiration))

        # Creates job to update gyms list
        if gyms_file is not None:
            gyms.load_from(gyms_file)
            self._scheduler.add_job(lambda: gyms.load_from(gyms_file), 'interval', hours=int(gyms_expiration))

        # Starts the scheduler
        self._scheduler.start()

        _LOGGER.info("Bot ready")

    def listen(self) -> None:
        _LOGGER.info("Start listening")

        # Begin to listen
        self._updater.start_polling()
        # Wait
        self._updater.idle()

    def _handler_error(self, update: Update, context: CallbackContext) -> None:
        _LOGGER.warning('Update "{}" caused error "{}"'.format(update, context.error))

    @Decorator.ChatMustBeEnabled
    def _handler_event_pinned(self, update: Update, _: CallbackContext) -> bool:
        # Check if the pin is caused by the bot
        if update.message.from_user.id != self._id:
            return False

        # Remove the notify message
        self._bot.delete_message(update.message.chat.id, update.message.message_id)

        return True

    @Decorator.ChatMustBeEnabled
    def _handler_screenshot(self, update: Update, _: CallbackContext) -> bool:
        _LOGGER.info("New image is arrived from {} by {}"
                     .format(update.effective_chat.title, update.effective_user.username))

        # Check if scan is disabled for this group
        if self._redis.exists(update.effective_chat.id):
            _LOGGER.info("Screenshots scan for chat {} is disabled".format(update.effective_chat.id))
            return False

        # Scan the screenshot
        self._scan_screenshot(update.message)
        return True

    @Decorator.ChatMustBeEnabled
    def _handler_set_hangout(self, update: Update, _: CallbackContext) -> bool:
        # Check if the reply is for the bot
        if update.message.reply_to_message.from_user.id != self._id:
            return False

        try:
            # Search the code in the bot message
            code = re.search(r"\[([a-zA-Z0-9]{8})\]", update.message.reply_to_message.text).group(1)
            # Try to retrieve the raid information
            raid = pickle.loads(self._redis.get(redis_keys.RAID.format(code)))
        except Exception:  # TODO: improve except
            traceback.print_exc()
            _LOGGER.warning("A invalid to bot message reply was come")
            return False

        _LOGGER.info("A reply to bot message was come")

        # Find the new hangout
        result = re.search(r"([0-2]?[0-9])[:.,]([0-5]?[0-9])", update.message.text)
        # Set new hangout
        raid.hangout = datetime.time(int(result.group(1)), int(result.group(2)))

        _LOGGER.debug(raid)

        # Save the raid in the db
        self._redis.setex(redis_keys.RAID.format(raid.code), 60 * 60 * 6, pickle.dumps(raid))

        # Try to delete user message
        self._try_to_delete(update.message)

        # Updates the message
        self._post_raid(raid, update.message.reply_to_message)

        return True

    @Decorator.ChatMustBeEnabled
    def _handler_buttons(self, update: Update, _: CallbackContext) -> bool:
        try:
            # Validate the data
            result = re.match(r"([a-zA-Z0-9]{8}):([arhif])", update.callback_query.data)
            # Try to retrieve the raid information
            raid = pickle.loads(self._redis.get(redis_keys.RAID.format(result.group(1))))
            # Get operation
            op = result.group(2)
        except Exception:  # TODO: improve except
            _LOGGER.warning("A invalid callback query was come")
            return False

        _LOGGER.info("A callback query was come")

        # Edit list of participants
        if op == "a":
            raid.add_participant(update.callback_query.from_user)
        elif op == "r":
            raid.remove_participant(update.callback_query.from_user)
        elif op == "h":
            raid.toggle_remote(update.callback_query.from_user)
        elif op == "i":
            raid.toggle_remote_invite(update.callback_query.from_user)
        elif op == "f":
            raid.toggle_flyer(update.callback_query.from_user)
        else:
            return False

        _LOGGER.debug(raid)

        # Save the raid in the db
        self._redis.setex(redis_keys.RAID.format(raid.code), 60 * 60 * 6, pickle.dumps(raid))

        # Updates the message
        self._post_raid(raid, update.callback_query.message)

        return True

    @Decorator.ChatMustBeEnabled
    def _handler_set_boss(self, update: Update, _: CallbackContext) -> bool:
        # Check if the reply is for the bot
        if update.message.reply_to_message.from_user.id != self._id:
            return False

        try:
            # Search the code in the bot message
            code = re.search(r"\[([a-zA-Z0-9]{8})\]", update.message.reply_to_message.text).group(1)
            # Try to retrieve the raid information
            raid = pickle.loads(self._redis.get(redis_keys.RAID.format(code)))
        except Exception:  # TODO: improve except
            _LOGGER.warning("A invalid to bot message reply was come")
            return False

        _LOGGER.info("A request to change boss was come from {}({}) by {}({})"
                     .format(update.effective_chat.title, update.effective_chat.id,
                             update.effective_user.username, update.effective_user.id))

        _LOGGER.info("The user suggested \"{}\"".format(update.message.text.strip()))

        # Get the suggested boss name
        name = update.message.text.strip()

        # Search the boss
        b = bosses.find(name, 0.8)

        # If the boss wasn't found reply with an error
        if b is None:
            update.message.reply_markdown("Sorry, but i don't know *{}*".format(name))
            _LOGGER.info("A valid boss wasn't found")
            return False

        _LOGGER.info("\"{}\" was found".format(b.name))

        # Set the new boss
        raid.boss = b

        # Save the raid in the db
        self._redis.setex(redis_keys.RAID.format(raid.code), 60 * 60 * 6, pickle.dumps(raid))

        _LOGGER.debug(raid)

        # Try to delete user message
        self._try_to_delete(update.message)

        # Updates the message
        self._post_raid(raid, update.message.reply_to_message)

        return True

    @Decorator.ChatMustBeEnabled
    @Decorator.UserMustBeAdmin
    def _handler_command_disablescan(self, update: Update, _: CallbackContext) -> bool:
        # Add current chat to the db of disabled scan
        self._redis.sadd(redis_keys.DISABLEDSCAN, update.message.chat.id)

        _LOGGER.info("Disable scan for chat {}".format(update.message.chat.id))
        update.message.chat.send_message("The scan now is disabled")

        return True

    @Decorator.ChatMustBeEnabled
    @Decorator.UserMustBeAdmin
    def _handler_command_enablescan(self, update: Update, _: CallbackContext) -> bool:
        # Remove current chat from the db of disabled scan
        self._redis.srem(redis_keys.DISABLEDSCAN, update.message.chat.id)

        _LOGGER.info("Enable scan for chat {}".format(update.message.chat.id))
        update.message.chat.send_message("The scan now is enabled")

        return True

    @Decorator.ChatMustBeEnabled
    def _handler_command_scan(self, update: Update, _: CallbackContext) -> bool:
        _LOGGER.info("Required scan from {} by {}".format(update.message.chat.id, update.message.from_user.id))

        # Check if it is a reply to screenshot
        if update.message.reply_to_message is None or len(update.message.reply_to_message.photo) == 0:
            update.message.reply_text("It must be a reply to a screenshot")
            _LOGGER.info("Invalid scan command")
            return False

        # Try to delete user command
        self._try_to_delete(update.message)

        try:
            # Scan the screenshot
            self._scan_screenshot(update.message.reply_to_message)
        except:
            traceback.print_exc()
            return False

        return True

    @Decorator.UserMustBeBotAdmin
    def _handler_command_addadmin(self, update: Update, _: CallbackContext) -> bool:
        _LOGGER.info("User {} try to add {} as bot admin".format(update.message.from_user.id,
                                                                 update.message.reply_to_message.from_user.id))

        # Check if the cited user is already a bot admin
        if self._redis.sismember(redis_keys.ADMIN, update.message.reply_to_message.from_user.id):
            _LOGGER.info("User {} is already a bot admin".format(update.message.reply_to_message.from_user.id))
            update.message.reply_markdown("[{}](tg://user?id={}) is already a bot admin"
                                          .format(update.message.reply_to_message.from_user.username,
                                                  update.message.reply_to_message.from_user.id))
            return False

        # Add cited user as bot admin
        self._redis.sadd(redis_keys.ADMIN, update.message.reply_to_message.from_user.id)
        _LOGGER.info("User {} is now a bot admin".format(update.message.reply_to_message.from_user.id))
        update.message.reply_markdown("[{}](tg://user?id={}) is now a bot admin"
                                      .format(update.message.reply_to_message.from_user.username,
                                              update.message.reply_to_message.from_user.id))

        return True

    @Decorator.UserMustBeBotAdmin
    def _handler_command_removeadmin(self, update: Update, _: CallbackContext) -> bool:
        _LOGGER.info("User {} try to remove {} as bot admin".format(update.message.from_user.id,
                                                                    update.message.reply_to_message.from_user.id))

        # Check if the mentioned user is the superadmin
        if self._redis.get(redis_keys.SUPERADMIN) == update.message.reply_to_message.from_user.id:
            _LOGGER.info("User {} is the superadmin".format(update.message.reply_to_message.from_user.id))
            update.message.reply_markdown("[{}](tg://user?id={}) is the superadmin and it cannot be removed"
                                          .format(update.message.reply_to_message.from_user.username,
                                                  update.message.reply_to_message.from_user.id))
            return False

        # Check if the cited user is not a bot admin
        if not self._redis.sismember(redis_keys.ADMIN, update.message.reply_to_message.from_user.id):
            _LOGGER.info("User {} is not a bot admin".format(update.message.reply_to_message.from_user.id))
            update.message.reply_markdown("[{}](tg://user?id={}) is not a bot admin"
                                          .format(update.message.reply_to_message.from_user.username,
                                                  update.message.reply_to_message.from_user.id))
            return False

        # Remove cited user as bot admin
        self._redis.srem(redis_keys.ADMIN, update.message.reply_to_message.from_user.id)
        _LOGGER.info("User {} is no longer a bot admin".format(update.message.reply_to_message.from_user.id))
        update.message.reply_markdown("[{}](tg://user?id={}) is no longer a bot admin"
                                      .format(update.message.reply_to_message.from_user.username,
                                              update.message.reply_to_message.from_user.id))

        return True

    @Decorator.UserMustBeBotAdmin
    def _handler_command_enablechat(self, update: Update, _: CallbackContext) -> bool:
        _LOGGER.info("Bot admin {} try to enable the chat {}".format(update.message.from_user.id,
                                                                     update.message.chat.id))

        # Check if this chat is already enabled
        if self._redis.sismember(redis_keys.ENABLEDCHAT, update.message.chat.id):
            _LOGGER.info("Chat {} is already enabled".format(update.message.chat.id))
            update.message.reply_markdown("This chat is already enabled")
            return False

        # Add this chat to the enabled
        self._redis.sadd(redis_keys.ENABLEDCHAT, update.message.chat.id)
        _LOGGER.info("Chat {} is now enabled".format(update.message.chat.id))
        update.message.reply_markdown("This chat is now enabled")

        return True

    @Decorator.UserMustBeBotAdmin
    def _handler_command_disablechat(self, update: Update, _: CallbackContext) -> bool:
        _LOGGER.info("Bot admin {} try to disable the chat {}".format(update.message.from_user.id,
                                                                      update.message.chat.id))

        # Check if this chat is not enabled
        if not self._redis.sismember(redis_keys.ENABLEDCHAT, update.message.chat.id):
            _LOGGER.info("Chat {} is not enabled".format(update.message.chat.id))
            update.message.reply_markdown("This chat is not enabled")
            return False

        # Remove this chat to the enabled
        self._redis.srem(redis_keys.ENABLEDCHAT, update.message.chat.id)
        _LOGGER.info("Chat {} is no longer enabled".format(update.message.chat.id))
        update.message.reply_markdown("This chat is no longer enabled")

        return True

    def _scan_screenshot(self, message: Message) -> None:
        # Get the highest resolution image
        img = message.photo[-1].get_file().download_as_bytearray()

        # Load the screenshot
        screen = ScreenshotRaid(img)

        # Check if it's a screenshot of a raid
        if not screen.is_raid:
            return

        _LOGGER.info("It's a valid screen of a raid")

        # Get the raid dataclass
        raid = screen.to_raid()

        # Save the raid in the db
        self._redis.setex(redis_keys.RAID.format(raid.code), 60 * 60 * 6, pickle.dumps(raid))

        # Send reply
        try:
            self._post_raid(raid, message)
        except:
            traceback.print_exc()  # TODO: Remove this debug method

        # Save sections of image if it is required
        if self._debug_folder is not None:
            try:
                os.makedirs(self._debug_folder, exist_ok=True)

                res = cv2.imwrite(os.path.join(self._debug_folder, "{}-anchors.png".format(raid.code)),
                                  screen._get_anchors_image())
                for s in screen._image_sections:
                    res = cv2.imwrite(os.path.join(self._debug_folder, "{}-{}.png".format(raid.code, s)),
                                      screen._image_sections[s]) and res
                if not res:
                    _LOGGER.warning("Something was gone wrong during save sections of image")

            except PermissionError:
                _LOGGER.warning("Unable to create debug folder")

    def _post_raid(self, raid: Raid, message: Message) -> None:
        options = {
            "disable_web_page_preview": True,
            "parse_mode": ParseMode.MARKDOWN_V2
        }

        # If the hangout is defined add the reply button to the message
        if raid.hangout is not None:
            options["reply_markup"] = InlineKeyboardMarkup([
                [
                    InlineKeyboardButton("\U00002795", callback_data=raid.code + ":a"),
                    InlineKeyboardButton("\U00002796", callback_data=raid.code + ":r")
                ],
                [
                    InlineKeyboardButton("\U0001F3E1", callback_data=raid.code + ":h"),
                    InlineKeyboardButton("\U0001F48C", callback_data=raid.code + ":i"),
                    InlineKeyboardButton("\U00002708", callback_data=raid.code + ":f")
                ]
            ])

        # If the reference message is a screenshot, the bot replies to that
        elif message.from_user.id != self._id:
            options["reply_to_message_id"] = message.message_id

        # TODO: improve this check method
        # Check if the old message was pinned
        try:
            pinned = self._bot.get_chat(message.chat.id).pinned_message.message_id == message.message_id
        except AttributeError:
            pinned = False

        # Try to delete the screenshot if it's necessary
        if message.reply_to_message is not None and raid.hangout is not None:
            self._try_to_delete(message.reply_to_message)

        # Try to delete the old message
        if message.from_user.id == self._id:
            self._try_to_delete(message)

        # Send new message
        new_msg = message.chat.send_message(raid.to_msg(),
                                            **options)

        # Re-pin the new message
        if pinned:
            self._bot.pin_chat_message(message.chat.id, new_msg.message_id, disable_notification=True)

    def _try_to_delete(self, message: Message):
        try:
            self._bot.delete_message(message.chat.id, message.message_id)
        except error.BadRequest:
            _LOGGER.info("The bot hasn't the permission to delete messages")

    def _init_db(self):
        pass



class ValueNotFound(Exception):
    pass


class ValueUnreadable(Exception):
    pass


class TimeNotFound(ValueNotFound):
    pass


class TimerNotFound(ValueNotFound):
    pass


class TimerUnreadable(ValueUnreadable):
    pass


class HatchingTimerException(Exception):
    pass


class HatchingTimerNotFound(HatchingTimerException, TimerNotFound):
    pass


class HatchingTimerUnreadable(HatchingTimerException, TimerUnreadable):
    pass


class RaidTimerException(Exception):
    pass


class RaidTimerNotFound(RaidTimerException, TimerNotFound):
    pass


class RaidTimerUnreadable(RaidTimerException, TimerUnreadable):
    pass


class ExTagException(Exception):
    pass


class ExTagNotFound(ExTagException, ValueNotFound):
    pass


class ExTagUnreadable(ExTagException, ValueUnreadable):
    pass


class GymNotFound(ValueNotFound):
    pass


class LevelNotFound(ValueNotFound):
    pass


class BossesListNotAvailable(Exception):
    pass


class BossNotFound(ValueNotFound):
    pass



from __future__ import annotations

import logging
from dataclasses import dataclass
from difflib import SequenceMatcher
from typing import List, Union
from urllib.parse import urlparse

import requests

from .exceptions import InvalidJSON, InvalidCSV

_LOGGER = logging.getLogger(__package__)


@dataclass
class Data:
    name: str


class DataList(List):
    def __init__(self):
        super(DataList, self).__init__()
        self._is_loaded = False

    @property
    def is_loaded(self) -> bool:
        return self._is_loaded

    def load_from(self, file: str) -> bool:
        _LOGGER.info("Try to load {}".format(self.__class__.__name__))

        try:
            # Check if the resource is remote
            if bool(urlparse(file).scheme):
                # Load the remote json
                raw = requests.get(file).text

            else:
                # Open the file and load it as json
                with open(file, 'r') as f:
                    raw = f.read()

        except FileNotFoundError:
            _LOGGER.warning("Failed to load the list: file not found")
            return False
        except requests.exceptions.ConnectionError:
            _LOGGER.warning("Failed to load the list: an HTTP error occurred")
            return False

        try:
            self._load_json(raw)
            self._is_loaded = True
        except (InvalidJSON, NotImplementedError):
            pass

        try:
            self._load_csv(raw)
            self._is_loaded = True
        except (InvalidCSV, NotImplementedError):
            pass

        if not self.is_loaded:
            _LOGGER.warning("The file is in a wrong format")
            return False

        _LOGGER.debug(self)

        _LOGGER.info("{} is loaded with {} entities".format(self.__class__.__name__, len(self)))

        return True

    def find(self, name: str, minimal_value: float = 0.4) -> Union[Data, None]:
        if not self.is_loaded:
            return None

        _LOGGER.debug("Try to find a candidate for '{}'".format(name))
        # Compare the boss_name with each boss in the list and find the most similar

        values = map(lambda x: (x, SequenceMatcher(None, name.lower(), x.name.lower()).ratio()), self)

        value = max(values, key=lambda x: x[1])

        if value[1] >= minimal_value:
            _LOGGER.debug("Found '{}' with confidence {:.3f}".format(value[0].name, value[1]))
            return value[0]
        else:
            _LOGGER.debug("No candidate found")
            return None

    def _load_json(self, raw):
        raise NotImplementedError

    def _load_csv(self, raw):
        raise NotImplementedError



import datetime
import logging
import re
from difflib import SequenceMatcher
from functools import reduce
from typing import Tuple, Union

import cv2
import numpy as np
import pytesseract

from . import resources
from ..cachedmethod import CachedMethod
from ..data import Boss, Gym, gyms, bosses
from ..exceptions import HatchingTimerNotFound, HatchingTimerUnreadable, RaidTimerNotFound, RaidTimerUnreadable, \
    ExTagNotFound, ExTagUnreadable, LevelNotFound, TimeNotFound, HatchingTimerException, RaidTimerException, \
    GymNotFound, ExTagException, BossNotFound, BossesListNotAvailable
from ..raid import Raid

Rect = Tuple[Tuple[int, int], Tuple[int, int]]

_LOGGER = logging.getLogger(__package__)


class ScreenshotRaid:
    debug = False

    def __init__(self, img: Union[np.ndarray, bytearray]):

        if isinstance(img, np.ndarray):
            self._img = img
        elif isinstance(img, bytearray):
            self._img = cv2.imdecode(np.asarray(img, dtype="uint8"), cv2.IMREAD_COLOR)
        else:
            raise Exception  # TODO: create adhoc exception

        if ScreenshotRaid.debug:
            self._image_sections = {}

        self._size = (len(self._img[0]), len(self._img))

        self._anchors = {}
        self._anchors_available = 0

        self._find_anchors()

    def _calc_subset(self, *subs) -> Rect:
        if len(subs) == 1 and isinstance(subs[0], tuple):
            subs = subs[0]
        elif len(subs) == 2:
            subs = list(subs)
        else:
            raise Exception

        points = [[None, None], [None, None]]

        for i in [0, 1]:
            # Amount of pixels centered
            if isinstance(subs[i], int):
                points[0][i] = round(self._size[i] / 2 - subs[i] / 2 - 1)
                points[1][i] = round(self._size[i] / 2 + subs[i] / 2 - 1)

            # Perceptual of pixels centered
            elif isinstance(subs[i], float):
                points[0][i] = round((self._size[i] - 1) / 2 - subs[i] * (self._size[i] - 1) / 2)
                points[1][i] = round((self._size[i] - 1) / 2 + subs[i] * (self._size[i] - 1) / 2)

            else:
                for j in [0, 1]:
                    v = subs[i][j]
                    if isinstance(v, float):
                        if v < 0:
                            points[j][i] = round((self._size[i] - 1) * v) + self._size[i]
                        else:
                            points[j][i] = round((self._size[i] - 1) * v)
                    else:
                        if v < 0:
                            points[j][i] = v + self._size[i]
                        else:
                            points[j][i] = v

        return (tuple(points[0]), tuple(points[1]))

    def _subset(self, rect: Rect) -> np.ndarray:
        return self._img[rect[0][1]:rect[1][1], rect[0][0]:rect[1][0]]

    def _read_hatching_timer(self) -> datetime.timedelta:
        if self.hatching_timer_position is None:
            raise HatchingTimerNotFound
        img = self._subset(self.hatching_timer_position)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        __, img = cv2.threshold(img, 210, 255, cv2.THRESH_BINARY_INV)

        if ScreenshotRaid.debug:
            self._image_sections["hatching_timer"] = img

        text = pytesseract.image_to_string(img, config=('--oem 1 --psm 3'))

        _LOGGER.debug("raw hatching_timer «{}»".format(text))

        result = re.search(r"([0-3])[:.]([0-5][0-9])[:.]([0-5][0-9])", text)

        try:
            _LOGGER.debug("hatching_timer {}:{}:{}".format(result.group(1), result.group(2), result.group(3)))
            return datetime.timedelta(hours=int(result.group(1)), minutes=int(result.group(2)),
                                      seconds=int(result.group(3)))
        except Exception:
            pass

        _LOGGER.debug("hatching timer unreadable")
        raise HatchingTimerUnreadable

    def _find_hatching_timer(self) -> Rect:
        red_lower = np.array([150, 100, 230])
        red_upper = np.array([255, 130, 255])

        sub = self._calc_subset(0.35, (0.15, 0.29))

        img = self._subset(sub)
        img = cv2.GaussianBlur(img, (5, 5), 5)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)

        mask = cv2.inRange(img, red_lower, red_upper)

        if ScreenshotRaid.debug:
            self._image_sections["hatching_timer_mask"] = mask

        try:
            contours, _ = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

            x, y, w, h = cv2.boundingRect(
                reduce((lambda x, y: x if cv2.contourArea(x) > cv2.contourArea(y) else y), contours))

            return ((sub[0][0] + x, sub[0][1] + y), (sub[0][0] + x + w, sub[0][1] + y + h))
        except:
            pass

        _LOGGER.debug("hatching timer not found")
        raise HatchingTimerNotFound

    def _read_raid_timer(self) -> datetime.timedelta:
        if self.raid_timer_position is None:
            raise RaidTimerNotFound
        img = self._subset(self.raid_timer_position)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        __, img = cv2.threshold(img, 210, 255, cv2.THRESH_BINARY_INV)

        if ScreenshotRaid.debug:
            self._image_sections["raid_timer"] = img

        text = pytesseract.image_to_string(img, config=('--oem 1 --psm 3'))

        _LOGGER.debug("raw raid_timer «{}»".format(text))

        result = re.search(r"([0-3])[:.]([0-5][0-9])[:.]([0-5][0-9])", text)

        try:
            _LOGGER.debug("raid_timer {}:{}:{}".format(result.group(1), result.group(2), result.group(3)))
            return datetime.timedelta(hours=int(result.group(1)), minutes=int(result.group(2)),
                                      seconds=int(result.group(3)))
        except Exception:
            pass

        _LOGGER.debug("raid timer unreadable")
        raise RaidTimerUnreadable

    def _find_raid_timer(self) -> Rect:
        red_lower = np.array([-50, 175, 230])
        red_upper = np.array([50, 205, 255])

        sub = self._calc_subset(((-0.30, -0.02), (0.54, 0.65)))

        img = self._subset(sub)
        img = cv2.GaussianBlur(img, (5, 5), 5)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)

        mask = cv2.inRange(img, red_lower, red_upper)

        if ScreenshotRaid.debug:
            self._image_sections["raid_timer_mask"] = mask

        try:
            contours, _ = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

            x, y, w, h = cv2.boundingRect(
                reduce((lambda x, y: x if cv2.contourArea(x) > cv2.contourArea(y) else y), contours))

            return ((sub[0][0] + x, sub[0][1] + y), (sub[0][0] + x + w, sub[0][1] + y + h))
        except:
            pass

        _LOGGER.debug("raid timer not found")
        raise RaidTimerNotFound

    def _find_gym(self) -> Gym:
        # TODO: improve find gym method
        try:
            x, y, r = self._anchors["gym_image"]

            sub = self._calc_subset((x + r + 10, -160), (y - r + 5, y + r - 5))
        except:
            sub = self._calc_subset((200, -160), (60, 150))

        img = self._subset(sub)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        __, img = cv2.threshold(img, 220, 255, cv2.THRESH_BINARY_INV)
        text = pytesseract.image_to_string(img, config=('--oem 1 --psm 3'))

        _LOGGER.debug("raw gym_name «{}»".format(text))

        text = text.rstrip().replace('\n', ' ')
        text = " ".join(text.split())

        logging.debug(text)

        if ScreenshotRaid.debug:
            self._image_sections["gym_name"] = img

        g = gyms.find(text)

        if g is not None:
            return g

        return Gym(name=text)
        # TODO: add the exception case

    def _find_boss(self) -> Union[Boss, None]:
        # Check if a list of available bosses was provided
        if not bosses.is_loaded:
            raise BossesListNotAvailable

        # Force the calc of the level if it isn't already calculated
        _ = self.level

        # Calculate the subset based on the level position
        try:
            (_, _), (_, y) = self._anchors["level"]
            sub = self._calc_subset(0.8, (y + 85, y + 215))
        except:
            sub = self._calc_subset(0.8, (0.23, 0.34))

        # Create the subset of the screenshot and filter it
        img = self._subset(sub)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        __, img = cv2.threshold(img, 240, 255, cv2.THRESH_BINARY_INV)

        if ScreenshotRaid.debug:
            self._image_sections["boss"] = img

        # Find the text in the subset
        text = pytesseract.image_to_string(img, config=('--oem 1 --psm 3'))

        _LOGGER.debug("raw boss «{}»".format(text))

        # Removed new line from the text
        text = text.rstrip().replace('\n', ' ')
        text = " ".join(text.split())

        # Try to find a boss from text
        b = bosses.find(text)

        # Check if a valid boss was found
        if b is None:
            raise BossNotFound

        return b

    def _find_ex_tag(self) -> Rect:
        # Prepare a range of colors to search the ex label in HSV color space
        color = np.array([133, 138, 189])
        color_offset = np.array([50, 15, 13])
        color_range = color - color_offset, color + color_offset

        # Calculate a subset of coordinates in the screenshot where the ex label could be
        sub = self._calc_subset((-230, 1.0), (30, 100))

        # Create the subset of the screenshot, filter it and move it in the HSV color space
        img = self._subset(sub)
        img = cv2.GaussianBlur(img, (5, 5), 5)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)

        # Create a mask of the image in the range of colors and dilate it with the aim of reduce the noise
        mask = cv2.inRange(img, *color_range)
        mask = cv2.dilate(mask, np.ones((5, 5), np.uint8))

        if ScreenshotRaid.debug:
            self._image_sections["ex_tag_mask"] = mask

        # Try to search contours of mask
        contours, _ = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

        if len(contours) > 0:
            # Search the biggest blob and calculate the bounding rectangle
            x, y, w, h = cv2.boundingRect(
                reduce((lambda x, y: x if cv2.contourArea(x) > cv2.contourArea(y) else y), contours))

            # Return the position of the bounding rectangle of the biggest blob
            return ((sub[0][0] + x, sub[0][1] + y), (sub[0][0] + x + w, sub[0][1] + y + h))

        # It wasn't found a candidate as ex label
        raise ExTagNotFound

    def _check_ex_tag(self) -> bool:
        # Check if it was found a candidate as ex label
        if self.ex_tag_position is None:
            raise ExTagNotFound

        # Get a sub image of the possible ex label and binarize it
        img = self._subset(self.ex_tag_position)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        __, img = cv2.threshold(img, 210, 255, cv2.THRESH_BINARY_INV)

        if ScreenshotRaid.debug:
            self._image_sections["ex_tag"] = img

        # Read the sub image
        text = pytesseract.image_to_string(img, config=('--oem 1 --psm 3'))

        _LOGGER.debug("raw ex_tag «{}»".format(text))

        # If it is enough similar to ex label the gym will be considered ex
        if max([SequenceMatcher(None, "ex raid", text.lower()).ratio(),
                SequenceMatcher(None, "raid ex", text.lower()).ratio()]) > 0.4:
            return True

        raise ExTagUnreadable

    def _find_level(self) -> int:
        # For eggs and hatched different parameters are used
        if self.is_egg:
            ht_pos = self.hatching_timer_position
            sub = self._calc_subset(0.55, (ht_pos[1][1] + 15, ht_pos[1][1] + 105))
            template = resources.LEVEL_EGG
            mask = resources.LEVEL_MASK_EGG
            threshold = 0.914
            matchf = lambda img: cv2.matchTemplate(img, template, cv2.TM_CCORR_NORMED, None, mask)
        else:
            sub = self._calc_subset(0.5, (0.10, 0.20))
            template = resources.LEVEL_HATCHED
            mask = resources.LEVEL_MASK_HATCHED
            threshold = 0.94
            matchf = lambda img: cv2.matchTemplate(img, template, cv2.TM_CCORR_NORMED, None, mask)

        # Get the subset image
        img = self._subset(sub)

        # Search match of the marker in the gray scale subset
        res = matchf(cv2.cvtColor(img, cv2.COLOR_BGR2GRAY))

        # Filter the results with a threshold
        loc = np.where(res >= threshold)

        # The point found must be merged
        # Number of recurrences of unique points
        occ = []
        # Contains unique points
        matches = []
        # Parse all the points found
        for a in np.dstack(loc[::-1])[0]:
            # Compare all the point with other unique points evaluate the distance
            for i, match in enumerate(matches):
                # If the point is almost near to the unique point merges that
                if np.linalg.norm(a - match) <= 10:
                    matches[i] = (match * occ[i] + a) // (occ[i] + 1)
                    occ[i] += 1
                    a = None
                    break
            # if the point is not almost near to another unique point mark it as unique point
            if a is not None:
                matches.append(a)
                occ.append(1)

        # Calculate the mean of vertical position of the points
        mean = np.mean([i[1] for i in matches]) if len(matches) > 0 else 0
        # Remove all the points that that are too far between themselves
        marker = list(filter(lambda x: abs(x[1] - mean) < 20, matches))

        w, h = template.shape[::-1]

        if ScreenshotRaid.debug:
            for pt in marker:
                cv2.rectangle(img, (pt[0], pt[1]), (pt[0] + w, pt[1] + h), (0, 0, 255), 2)
            self._image_sections["level"] = img

        # Count the matches to calculate the level
        level = len(matches) if len(matches) < 5 else 5

        # No matches raises a exception
        if level == 0:
            raise LevelNotFound

        # Save the anchor
        self._anchors["level"] = (
            (
                min([pt[0] for pt in marker]) + sub[0][0],
                min([pt[1] for pt in marker]) + sub[0][1]
            ),
            (
                max([pt[0] for pt in marker]) + w + sub[0][0],
                max([pt[1] for pt in marker]) + h + sub[0][1]
            )
        )

        return level

    def _find_time(self) -> datetime.time:
        try:
            gym_image = self._anchors["gym_image"]
            ym = gym_image[1] - gym_image[2] - 10
        except:
            ym = 60

        for x in [
            0.2,
            (-0.2, 1.0),
            1.0
        ]:
            sub = self._calc_subset(x, (0, ym))
            img = self._subset(sub)
            img = cv2.GaussianBlur(img, (5, 5), 3)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 11, 2)

            if ScreenshotRaid.debug:
                self._image_sections["time"] = img

            text = pytesseract.image_to_string(img, config=('--oem 1 --psm 3'))

            logging.debug(text)

            result = re.search(r"([0-2]?[0-9]):([0-5][0-9])", text)

            try:
                return datetime.time(int(result.group(1)), int(result.group(2)))
            except:
                pass

        raise TimeNotFound

    def _find_anchors(self) -> None:
        ANCHORS = {
            "gym_image": [
                ((0, 0.25), (40, 0.20)),
                {"param1": 50, "param2": 30, "minRadius": 50, "maxRadius": 65}
            ],
            "gym_detail": [
                ((-0.20, -30), (60, 0.17)),
                {"param1": 50, "param2": 30, "minRadius": 20, "maxRadius": 40}
            ],
            "raid_info": [
                ((30, 0.25), (-250, 1.0)),
                {"param1": 50, "param2": 30, "minRadius": 30, "maxRadius": 40}
            ],
            "exit": [
                (0.25, (-250, 1.0)),
                {"param1": 50, "param2": 30, "minRadius": 30, "maxRadius": 40}
            ],
            "gym": [
                ((-0.25, -30), (-250, 1.0)),
                {"param1": 50, "param2": 30, "minRadius": 30, "maxRadius": 40}
            ]
        }

        # Doesn't search anchors are already found
        for a in self._anchors:
            ANCHORS.pop(a, None)

        for a in ANCHORS:
            sub = self._calc_subset(ANCHORS[a][0])
            img = self._subset(sub)
            circles = cv2.HoughCircles(cv2.cvtColor(img, cv2.COLOR_BGR2GRAY), cv2.HOUGH_GRADIENT, 1.2, 100,
                                       **ANCHORS[a][1])
            if circles is None:
                self._anchors[a] = None
                continue

            x, y, r = np.round(circles[0]).astype("int")[0]

            x += sub[0][0]
            y += sub[0][1]

            self._anchors_available += 1
            self._anchors[a] = (x, y, r)

    @property
    @CachedMethod
    def hatching_timer_position(self) -> Union[Rect, None]:
        try:
            self._anchors["hatching_timer"] = self._find_hatching_timer()
            return self._anchors["hatching_timer"]
        except HatchingTimerNotFound:
            return None

    @property
    @CachedMethod
    def raid_timer_position(self) -> Union[Rect, None]:
        try:
            self._anchors["raid_timer"] = self._find_raid_timer()
            return self._anchors["raid_timer"]
        except RaidTimerNotFound:
            return None

    @property
    @CachedMethod
    def ex_tag_position(self) -> Union[Rect, None]:
        try:
            return self._find_ex_tag()
        except ExTagNotFound:
            return None

    @property
    @CachedMethod
    def timer(self) -> Union[datetime.timedelta, None]:
        return self.hatching_timer if self.is_egg else self.raid_timer

    @property
    @CachedMethod
    def hatching_timer(self) -> Union[datetime.timedelta, None]:
        try:
            return self._read_hatching_timer()
        except HatchingTimerException:
            return None

    @property
    @CachedMethod
    def raid_timer(self) -> Union[datetime.timedelta, None]:
        try:
            return self._read_raid_timer()
        except RaidTimerException:
            return None

    @property
    @CachedMethod
    def gym(self) -> Union[Gym, None]:
        try:
            return self._find_gym()
        except GymNotFound:
            return None

    @property
    @CachedMethod
    def boss(self) -> Union[Boss, None]:
        try:
            return self._find_boss()
        except (BossesListNotAvailable, BossNotFound):
            return None

    @property
    @CachedMethod
    def is_ex(self) -> bool:
        try:
            if self._check_ex_tag():
                self._anchors["ex_tag"] = self.ex_tag_position
                return True
        except ExTagException:
            return False

    @property
    @CachedMethod
    def level(self) -> Union[int, None]:
        try:
            return self._find_level()
        except LevelNotFound:
            return None

    @property
    @CachedMethod
    def time(self) -> Union[datetime.time, None]:
        try:
            return None
            # The calc of the time from the notification bar is momentary disabled
            # return self._find_time()
        except TimeNotFound:
            return None

    @property
    @CachedMethod
    def is_raid(self) -> bool:
        # If there are not at least 4 anchors the screenshot is not a raid
        if self._anchors_available < 4:
            return False

        # Try to find the hatching timer
        if self.hatching_timer_position is not None:
            return True

        # Try to find the raid timer
        if self.raid_timer_position is not None:
            return True

        # If there is neither hatching timer nor raid timer then the screenshot is not a raid
        return False

    @property
    @CachedMethod
    def is_egg(self) -> bool:
        return self.hatching_timer is not None

    @property
    @CachedMethod
    def is_hatched(self) -> bool:
        return self.hatching_timer is None

    @property
    @CachedMethod
    def hatching(self) -> Union[datetime.time, None]:
        if self.hatching_timer is None:
            return None

        if self.time is not None:
            return (datetime.datetime.combine(datetime.date(1970, 1, 1), self.time) + self.hatching_timer).time()
        else:
            return (datetime.datetime.now() + self.hatching_timer).time()

    @property
    @CachedMethod
    def end(self) -> Union[datetime.time, None]:
        # Get time or use system time
        time = datetime.datetime.combine(datetime.date(1970, 1, 1), self.time) \
            if self.time is not None else datetime.datetime.now()

        if self.is_hatched:
            if self.raid_timer is not None:
                return (time + self.raid_timer).time()
            else:
                return None
        else:
            if self.hatching_timer is not None:
                return (time + self.hatching_timer + datetime.timedelta(minutes=45)).time()
            else:
                return None

    def to_raid(self) -> Raid:
        if self.is_hatched:
            return Raid(gym=self.gym,
                        level=self.level,
                        end=self.end,
                        boss=self.boss,
                        is_hatched=True,
                        is_ex=self.is_ex,
                        is_aprx_time=(True if self.time is None else False))
        else:
            return Raid(gym=self.gym,
                        level=self.level,
                        hatching=self.hatching,
                        end=self.end,
                        is_hatched=False,
                        is_ex=self.is_ex,
                        is_aprx_time=(True if self.time is None else False))

    def _get_anchors_image(self) -> np.ndarray:
        img = self._img.copy()
        for i in self._anchors.values():
            try:
                if len(i) == 3:
                    cv2.circle(img, (i[0], i[1]), i[2], (0, 255, 0), 2)
                    cv2.circle(img, (i[0], i[1]), 2, (0, 0, 255), 3)
                elif len(i) == 2:
                    cv2.rectangle(img, i[0], i[1], (0, 255, 0), 2)
            except Exception:
                pass

        return img



import logging
import pkgutil
from typing import Union

import pogoraidbot


def logger_setup(log_level: Union[int, str] = logging.ERROR, modules_log_level: int = logging.ERROR):
    # sets format for the log
    logging.basicConfig(format="%(levelname)s|%(name)s|%(message)s", level=modules_log_level)

    # sets log level for local packages
    for _, modname, _ in pkgutil.walk_packages(path=pogoraidbot.__path__,
                                               prefix=pogoraidbot.__name__ + ".",
                                               onerror=lambda x: None):
        logging.getLogger(modname).setLevel(log_level)



from setuptools import setup, find_packages

__version__ = "0.0"
exec(open("./pogoraidbot/version.py").read())

with open('README.md') as f:
    _LONG_DESCRIPTION = f.read()

setup(
    name='pogoraidbot',
    packages=find_packages(),
    version=__version__,
    license='gpl-3.0',
    description='A telegram bot to organize PoGo raid that it can be self hosted',
    long_description=_LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Roberto Bochet',
    author_email='robertobochet@gmail.com',
    url='https://github.com/RobertoBochet/pogoraidbot',
    keywords=['game', 'pokemongo', 'pogo', 'pokemongo-raid', 'telegram', 'telegrambot'],
    install_requires=[
        'python-telegram-bot ~= 12.7',
        'opencv-python ~= 4.1',
        'pytesseract ~= 0.3',
        'redis ~= 3.5',
        'requests ~= 2.22',
        'schema ~= 0.7',
        'apscheduler ~= 3.6',
        'mpu ~= 0.23'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.6'
)



from __future__ import annotations

import csv
import json
from dataclasses import dataclass
from io import StringIO

from mpu.string import str2bool
from schema import Schema, Or, Optional

from .data import Data, DataList
from .exceptions import InvalidJSON, InvalidCSV


@dataclass
class Boss(Data):
    level: int = None
    is_there_shiny: bool = False


class BossesList(DataList):
    def _load_json(self, raw: str) -> None:
        try:
            data = json.loads(raw)
        except ValueError:
            raise InvalidJSON

        # Simple list of boss' names
        schema1 = Schema([str])

        # Dictionary of boss' names and level
        schema2 = Schema({
            str: Or(1, 2, 3, 4, 5)
        })

        # List of Boss objects
        schema3 = Schema([{
            "name": str,
            Optional("level"): int,
            Optional("is_there_shiny"): bool
        }])

        # Check if the list is valid
        if schema1.is_valid(data):
            self.clear()
            for b in data:
                self.append(Boss(b))
            return

        elif schema2.is_valid(data):
            self.clear()
            for b in data:
                self.append(Boss(
                    name=b,
                    level=data[b]
                ))
            return

        elif schema3.is_valid(data):
            self.clear()
            for b in data:
                self.append(Boss(**b))
            return

        raise InvalidJSON

    def _load_csv(self, raw: str) -> None:

        rows = csv.reader(StringIO(raw), skipinitialspace=True)

        # skips the header row and counts the columns
        c = len(next(rows))

        if c == 1:
            for row in rows:
                self.append(Boss(row[0].strip()))
            return

        elif c == 2:
            for row in rows:
                try:
                    if int(row[1].strip()) < 0 or int(row[1].strip()) > 5:
                        raise InvalidCSV
                except:
                    raise InvalidCSV

                self.append(Boss(row[0].strip(), int(row[1].strip())))
            return

        elif c == 3:
            for row in rows:
                try:
                    if int(row[1].strip()) < 0 or int(row[1].strip()) > 5:
                        raise InvalidCSV
                    str2bool(row[2].strip())
                except:
                    raise InvalidCSV

                self.append(Boss(row[0].strip(), int(row[1].strip()), str2bool(row[2].strip())))
            return

        raise InvalidCSV
